@@ -7,11 +7,10 @@ import (
 	"strings"
 	"time"
 
-	workv1client "github.com/open-cluster-management/api/client/work/clientset/versioned/typed/work/v1"
-	workapiv1 "github.com/open-cluster-management/api/work/v1"
 	"github.com/openshift/library-go/pkg/controller/factory"
 	"github.com/openshift/library-go/pkg/operator/events"
 	"github.com/openshift/library-go/pkg/operator/resource/resourcehelper"
+	"github.com/openshift/library-go/pkg/operator/resource/resourcemerge"
 	apiextensionsv1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1"
 	apiextensionsv1beta1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1beta1"
 	"k8s.io/apimachinery/pkg/api/equality"
@@ -24,6 +23,8 @@ import (
 	"k8s.io/client-go/dynamic"
 	"k8s.io/client-go/util/retry"
 	"k8s.io/klog/v2"
+	workv1client "open-cluster-management.io/api/client/work/clientset/versioned/typed/work/v1"
+	workapiv1 "open-cluster-management.io/api/work/v1"
 )
 
 const (
@@ -138,48 +139,69 @@ type UpdateManifestWorkStatusFunc func(status *workapiv1.ManifestWorkStatus) err
 func UpdateManifestWorkStatus(
 	ctx context.Context,
 	client workv1client.ManifestWorkInterface,
-	workName string,
+	manifestWork *workapiv1.ManifestWork,
 	updateFuncs ...UpdateManifestWorkStatusFunc) (*workapiv1.ManifestWorkStatus, bool, error) {
-	updated := false
-	var updatedWorkStatus *workapiv1.ManifestWorkStatus
-	err := retry.RetryOnConflict(retry.DefaultBackoff, func() error {
-		manifestWork, err := client.Get(ctx, workName, metav1.GetOptions{})
+	// in order to reduce the number of GET requests to hub apiserver, try to update the manifestwork
+	// fetched from informer cache (with lister).
+	updatedWorkStatus, updated, err := updateManifestWorkStatus(ctx, client, manifestWork, updateFuncs...)
+	if err == nil {
+		return updatedWorkStatus, updated, nil
+	}
+
+	// if the update failed, retry with the manifestwork resource fetched with work client.
+	err = retry.RetryOnConflict(retry.DefaultBackoff, func() error {
+		manifestWork, err := client.Get(ctx, manifestWork.Name, metav1.GetOptions{})
 		if err != nil {
 			return err
 		}
-		oldStatus := &manifestWork.Status
-
-		newStatus := oldStatus.DeepCopy()
-		for _, update := range updateFuncs {
-			if err := update(newStatus); err != nil {
-				return err
-			}
-		}
-		if equality.Semantic.DeepEqual(oldStatus, newStatus) {
-			// We return the newStatus which is a deep copy of oldStatus but with all update funcs applied.
-			updatedWorkStatus = newStatus
-			return nil
-		}
-
-		manifestWork.Status = *newStatus
-		updatedManifestWork, err := client.UpdateStatus(ctx, manifestWork, metav1.UpdateOptions{})
-		if err != nil {
-			return err
-		}
-		updatedWorkStatus = &updatedManifestWork.Status
-		updated = err == nil
+		updatedWorkStatus, updated, err = updateManifestWorkStatus(ctx, client, manifestWork, updateFuncs...)
 		return err
 	})
 
 	return updatedWorkStatus, updated, err
 }
 
+// updateManifestWorkStatus updates the status of the given manifestWork. The manifestWork is mutated.
+func updateManifestWorkStatus(
+	ctx context.Context,
+	client workv1client.ManifestWorkInterface,
+	manifestWork *workapiv1.ManifestWork,
+	updateFuncs ...UpdateManifestWorkStatusFunc) (*workapiv1.ManifestWorkStatus, bool, error) {
+	oldStatus := &manifestWork.Status
+	newStatus := oldStatus.DeepCopy()
+	for _, update := range updateFuncs {
+		if err := update(newStatus); err != nil {
+			return nil, false, err
+		}
+	}
+	if equality.Semantic.DeepEqual(oldStatus, newStatus) {
+		// We return the newStatus which is a deep copy of oldStatus but with all update funcs applied.
+		return newStatus, false, nil
+	}
+
+	manifestWork.Status = *newStatus
+	updatedManifestWork, err := client.UpdateStatus(ctx, manifestWork, metav1.UpdateOptions{})
+	if err != nil {
+		return nil, false, err
+	}
+	return &updatedManifestWork.Status, true, nil
+}
+
 // DeleteAppliedResources deletes all given applied resources and returns those pending for finalization
 // If the uid recorded in resources is different from what we get by client, ignore the deletion.
-func DeleteAppliedResources(resources []workapiv1.AppliedManifestResourceMeta, reason string, dynamicClient dynamic.Interface, recorder events.Recorder) ([]workapiv1.AppliedManifestResourceMeta, []error) {
+func DeleteAppliedResources(
+	resources []workapiv1.AppliedManifestResourceMeta,
+	reason string,
+	dynamicClient dynamic.Interface,
+	recorder events.Recorder,
+	owner metav1.OwnerReference) ([]workapiv1.AppliedManifestResourceMeta, []error) {
 	var resourcesPendingFinalization []workapiv1.AppliedManifestResourceMeta
 	var errs []error
 
+	// set owner to be removed
+	ownerCopy := owner.DeepCopy()
+	ownerCopy.UID = types.UID(fmt.Sprintf("%s-", owner.UID))
+
 	for _, resource := range resources {
 		gvr := schema.GroupVersionResource{Group: resource.Group, Version: resource.Version, Resource: resource.Resource}
 		u, err := dynamicClient.
@@ -198,6 +220,34 @@ func DeleteAppliedResources(resources []workapiv1.AppliedManifestResourceMeta, r
 			continue
 		}
 
+		existingOwner := u.GetOwnerReferences()
+
+		// If it is not owned by us, skip
+		if !IsOwnedBy(owner, existingOwner) {
+			continue
+		}
+
+		// Merge with the existing owners to move the owner.
+		modified := resourcemerge.BoolPtr(false)
+		resourcemerge.MergeOwnerRefs(modified, &existingOwner, []metav1.OwnerReference{*ownerCopy})
+
+		// If there are still any other existing owners (not only ManifestWorks), update ownerrefs only.
+		if len(existingOwner) > 0 {
+			if !*modified {
+				continue
+			}
+
+			u.SetOwnerReferences(existingOwner)
+			_, err = dynamicClient.Resource(gvr).Namespace(resource.Namespace).Update(context.TODO(), u, metav1.UpdateOptions{})
+			if err != nil {
+				errs = append(errs, fmt.Errorf(
+					"Failed to remove owner from resource %v with key %s/%s: %w",
+					gvr, resource.Namespace, resource.Name, err))
+			}
+
+			continue
+		}
+
 		if resource.UID != string(u.GetUID()) {
 			// the traced instance has been deleted, and forget this item.
 			continue
@@ -276,6 +326,7 @@ func AppliedManifestworkQueueKeyFunc(hubhash string) factory.ObjectQueueKeyFunc
 		if !strings.HasPrefix(accessor.GetName(), hubhash) {
 			return ""
 		}
+
 		return strings.TrimPrefix(accessor.GetName(), hubhash+"-")
 	}
 }
@@ -285,3 +336,22 @@ func AppliedManifestworkQueueKeyFunc(hubhash string) factory.ObjectQueueKeyFunc
 func HubHash(hubServer string) string {
 	return fmt.Sprintf("%x", sha256.Sum256([]byte(hubServer)))
 }
+
+// IsOwnedBy check if owner exists in the ownerrefs.
+func IsOwnedBy(myOwner metav1.OwnerReference, existingOwners []metav1.OwnerReference) bool {
+	for _, owner := range existingOwners {
+		if myOwner.UID == owner.UID {
+			return true
+		}
+	}
+	return false
+}
+
+func NewAppliedManifestWorkOwner(appliedWork *workapiv1.AppliedManifestWork) *metav1.OwnerReference {
+	return &metav1.OwnerReference{
+		APIVersion: workapiv1.GroupVersion.WithKind("AppliedManifestWork").GroupVersion().String(),
+		Kind:       workapiv1.GroupVersion.WithKind("AppliedManifestWork").Kind,
+		Name:       appliedWork.Name,
+		UID:        appliedWork.UID,
+	}
+}
